#if __linux__ || __APPLE__
#define memspeed_unix 1
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>

void consolestring( const char* str ) {
    printf( str );
}

typedef void* MemoryRegion;
typedef int Error;
typedef caddr_t Address;
#define MemoryAllocationFailed -1
#define Success 0

struct timeval start, stop;
void StartTimer() {
    gettimeofday( &start, NULL );
}
void StopTimer() {
    gettimeofday( &stop, NULL );
}
unsigned long BenchTime() {
    return (stop.tv_sec - start.tv_sec) * 1000000 + stop.tv_usec - start.tv_usec;
}
#define TICKS_TO_USEC( n ) ( n )
const char* ErrorString( int a ) {
    return "";
}

#ifndef MAX_PAGES
#define MAX_PAGES 512
#endif
#else
#include <INTEGRITY.h>
#include <memory_region.h>
#include "bench.h"
#endif

static void my_printf(const char *fmt, ...)
{
    static char buf[256];
    va_list ap;
    va_start(ap, fmt);
    vsprintf(buf, fmt, ap);
    consolestring(buf); /* displays to serial port if supported */
    va_end(ap);
}

//extern MemoryRegionPool* __ghs_VirtualMemoryRegionPool;

#ifndef verbose
int verbose = 0;
#endif

/* (T)hrow (E)rror */
#define TE( n ) { Error __te_e = ( n ); if ( __te_e != Success ) { printf("%s:%d \"%s\" = %s\n", __FILE__, __LINE__, #n, ErrorString( __te_e ) ); return __te_e; } else if ( verbose ) { printf( "%4d %s\n", __LINE__, #n );} }

struct page {
    MemoryRegion pmr;
    MemoryRegion vmr;
    Address vFirst;//, vLast;
    unsigned long bytes;
    struct page* next;
};
typedef struct page Page;

#if memspeed_unix
#ifndef PAGE_SIZE
#define PAGE_SIZE (4*1024)
#endif
Error getPage( Page** r ) {
    Page* nr;
    
    nr = (Page*)malloc(sizeof(Page));
    nr->pmr = nr->vmr = malloc( PAGE_SIZE );
    nr->vFirst = (Address)nr->vmr;
    nr->bytes = PAGE_SIZE;
    nr->next = *r;
    *r = nr;
    return 0;
}
#else
int PAGE_SIZE = 4096;
Error getPage( Page** r ) {
    Page tp;
    Page* nr;
    Error e;
    e = GetPageFromAddressSpaceFreeList( PrimaryAddressSpace, &(tp.pmr) );
    if ( e == Success ) {
    } else if ( e == NoPagesOnFreeList ) {
	return e;
    } else {
	TE(e);
    }
    if ( __ghs_VirtualMemoryRegionPool ) {
    TE(MapMemory( __ghs_VirtualMemoryRegionPool, tp.pmr, &(tp.vmr), &(tp.vFirst), &(tp.bytes) ));
    } else {
	Address vLast;
	tp.vmr = tp.pmr;
	TE(GetMemoryRegionAddresses( tp.vmr, &(tp.vFirst), &(vLast) ));
	tp.bytes = vLast - tp.vFirst + 1;
    }
	PAGE_SIZE = tp.bytes;
    nr = (Page*)malloc(sizeof(Page));
    *nr = tp;
    nr->next = *r;
    *r = nr;
    return Success;
}
#endif

// the target of throwaway operations to make sure reads/writes happen
int globaldummy;

/* Store results so that background printing doesn't slow down the test.
   Print all at end. */
struct result {
    const char* label;
    int repeat;
    int pages;
    unsigned long bytes;
    unsigned long usec;
    long long Bps;
    struct result* next;
};
typedef struct result Result;
Result* results = NULL;

Error newResult( const char* label, int repeat, int pages, unsigned long bytes, unsigned long usec, long long Bps ) {
    Result* r;
    r = (Result*)malloc(sizeof(Result));
    if ( r == NULL ) {
	return MemoryAllocationFailed;
    }
    r->label = label;
    r->repeat = repeat;
    r->pages = pages;
    r->bytes = bytes;
    r->usec = usec;
    r->Bps = Bps;
    r->next = results;
    results = r;
    return Success;
}
void reverseResults(void) {
    Result* nl = NULL;
    Result* next;
    while ( results ) {
	next = results->next;
	results->next = nl;
	nl = results;
	results = next;
    }
    results = nl;
}
void printResult( Result* r ) {
    my_printf("%s\trepeat=%d\tpages=%d\tbytes=%d\tusec=%d\tBps=%lld\n", r->label, r->repeat, r->pages * PAGE_SIZE, r->bytes, r->usec, r->Bps );
}
void printAllResults(void) {
    Result* r;
    reverseResults();
    r = results;
    while ( r ) {
#if ! memspeed_unix
	Time t = { 0, 0xffffffff / 100 };
	Sleep( &t );
#endif
	printResult( r );
	r = r->next;
    }
}

Error testReadSpeed1( Page* pIn, int repeat ) {
    int pages = 0;
    unsigned long bytes = 0;
    int dummy = 0;
    Address vf;
    Page* p;
    int repeatNotDone = repeat;
    p = pIn;
    while ( p ) {
	pages++;
	p = p->next;
    }
    StartTimer();
    while ( repeatNotDone ) {
	p = pIn;
	repeatNotDone--;
	while ( p ) {
	    unsigned long len, i;
	    bytes += (len = p->bytes);
	    vf = p->vFirst;
	    for ( i = 0; i < len; i++ ) {
		dummy += *((volatile u_int8_t*)(vf + i));
	    }
	    p = p->next;
	}
    }
    StopTimer();
    {
	long long Bps;
	Bps = bytes;
	Bps *= 1000000;
	Bps /= TICKS_TO_USEC(BenchTime());
	TE(newResult("read1", repeat, pages, bytes, TICKS_TO_USEC(BenchTime()), Bps ));
    }
    globaldummy += dummy;
    return Success;
}

Error testReadSpeed2( Page* pIn, int repeat ) {
    int pages = 0;
    unsigned long bytes = 0;
    int dummy = 0;
    Address vf;
    Page* p;
    int repeatNotDone = repeat;
    p = pIn;
    while ( p ) {
	pages++;
	p = p->next;
    }
    StartTimer();
    while ( repeatNotDone ) {
	p = pIn;
	repeatNotDone--;
	while ( p ) {
	    unsigned long len, i;
	    bytes += (len = p->bytes);
	    vf = p->vFirst;
	    len = len >> 1;
	    for ( i = 0; i < len; i++ ) {
		dummy += *((volatile u_int16_t*)(vf + i));
	    }
	    p = p->next;
	}
    }
    StopTimer();
    {
	long long Bps;
	Bps = bytes;
	Bps *= 1000000;
	Bps /= TICKS_TO_USEC(BenchTime());
	TE(newResult("read2", repeat, pages, bytes, TICKS_TO_USEC(BenchTime()), Bps ));
    }
    globaldummy += dummy;
    return Success;
}

Error testReadSpeed4( Page* pIn, int repeat ) {
    int pages = 0;
    unsigned long bytes = 0;
    int dummy = 0;
    Address vf;
    Page* p;
    int repeatNotDone = repeat;
    p = pIn;
    while ( p ) {
	pages++;
	p = p->next;
    }
    StartTimer();
    while ( repeatNotDone ) {
	p = pIn;
	repeatNotDone--;
	while ( p ) {
	    unsigned long len, i;
	    bytes += (len = p->bytes);
	    vf = p->vFirst;
	    len = len >> 2;
	    for ( i = 0; i < len; i++ ) {
		dummy += *((volatile u_int32_t*)(vf + i));
	    }
	    p = p->next;
	}
    }
    StopTimer();
    {
	long long Bps;
	Bps = bytes;
	Bps *= 1000000;
	Bps /= TICKS_TO_USEC(BenchTime());
	TE(newResult("read4", repeat, pages, bytes, TICKS_TO_USEC(BenchTime()), Bps ));
    }
    globaldummy += dummy;
    return Success;
}

Error testWriteSpeed1( Page* pIn, int repeat ) {
    int pages = 0;
    unsigned long bytes = 0;
    Address vf;
    Page* p;
    int repeatNotDone = repeat;
    p = pIn;
    while ( p ) {
	pages++;
	p = p->next;
    }
    StartTimer();
    while ( repeatNotDone ) {
	p = pIn;
	repeatNotDone--;
	while ( p ) {
	    unsigned long len, i;
	    bytes += (len = p->bytes);
	    vf = p->vFirst;
	    for ( i = 0; i < len; i++ ) {
		*((volatile u_int8_t*)(vf + i)) = i & 0xff;
	    }
	    p = p->next;
	}
    }
    StopTimer();
    {
	long long Bps;
	Bps = bytes;
	Bps *= 1000000;
	Bps /= TICKS_TO_USEC(BenchTime());
	TE(newResult("write1", repeat, pages, bytes, TICKS_TO_USEC(BenchTime()), Bps ));
    }
    return Success;
}

Error testWriteSpeed2( Page* pIn, int repeat ) {
    int pages = 0;
    unsigned long bytes = 0;
    Address vf;
    Page* p;
    int repeatNotDone = repeat;
    p = pIn;
    while ( p ) {
	pages++;
	p = p->next;
    }
    StartTimer();
    while ( repeatNotDone ) {
	p = pIn;
	repeatNotDone--;
	while ( p ) {
	    unsigned long len, i;
	    bytes += (len = p->bytes);
	    vf = p->vFirst;
	    len = len >> 1;
	    for ( i = 0; i < len; i++ ) {
		*((volatile u_int16_t*)(vf + i)) = i & 0xffff;
	    }
	    p = p->next;
	}
    }
    StopTimer();
    {
	long long Bps;
	Bps = bytes;
	Bps *= 1000000;
	Bps /= TICKS_TO_USEC(BenchTime());
	TE(newResult("write2", repeat, pages, bytes, TICKS_TO_USEC(BenchTime()), Bps ));
    }
    return Success;
}

Error testWriteSpeed4( Page* pIn, int repeat ) {
    int pages = 0;
    unsigned long bytes = 0;
    Address vf;
    Page* p;
    int repeatNotDone = repeat;
    p = pIn;
    while ( p ) {
	pages++;
	p = p->next;
    }
    StartTimer();
    while ( repeatNotDone ) {
	p = pIn;
	repeatNotDone--;
	while ( p ) {
	    unsigned long len, i;
	    bytes += (len = p->bytes);
	    vf = p->vFirst;
	    len = len >> 2;
	    for ( i = 0; i < len; i++ ) {
		*((volatile u_int32_t*)(vf + i)) = i & 0xffffffff;
	    }
	    p = p->next;
	}
    }
    StopTimer();
    {
	long long Bps;
	Bps = bytes;
	Bps *= 1000000;
	Bps /= TICKS_TO_USEC(BenchTime());
	TE(newResult("write4", repeat, pages, bytes, TICKS_TO_USEC(BenchTime()), Bps ));
    }
    return Success;
}

Error testWriteLatency( Page* pIn, int repeat ) {
    int pages = 0;
    unsigned long bytes = 0;
    Page* p;
    int repeatNotDone;
    Address* addrs;
    int i, dummy = 0;
    repeat = repeat * 256;
    p = pIn;
    while ( p ) {
	pages++;
	p = p->next;
    }
    addrs = (Address*)malloc(sizeof(Address)*pages);
    if ( addrs == NULL ) {
	return MemoryAllocationFailed;
    }
    p = pIn;
    i = 0;
    while ( p ) {
	addrs[i] = p->vFirst;
	p = p->next;
	i++;
    }
    bytes = pages * repeat;
    repeatNotDone = repeat;
    StartTimer();
    while ( repeatNotDone ) {
	repeatNotDone--;
	for ( i = 0; i < pages; i++ ) {
	    *((volatile u_int32_t*)(addrs[i])) = (++dummy) & 0xffffffff;
	}
    }
    StopTimer();
    free( addrs );
    {
	long long Bps;
	Bps = bytes;
	Bps *= 1000000;
	Bps /= TICKS_TO_USEC(BenchTime());
	TE(newResult("writeL", repeat, pages, bytes, TICKS_TO_USEC(BenchTime()), Bps ));
    }
    return Success;
}

Error testReadLatency( Page* pIn, int repeat ) {
    int pages = 0;
    unsigned long bytes = 0;
    Page* p;
    int repeatNotDone;
    Address* addrs;
    int i, dummy = 0;
    repeat = repeat * 256;
    p = pIn;
    while ( p ) {
	pages++;
	p = p->next;
    }
    addrs = (Address*)malloc(sizeof(Address)*pages);
    if ( addrs == NULL ) {
	return MemoryAllocationFailed;
    }
    p = pIn;
    i = 0;
    while ( p ) {
	addrs[i] = p->vFirst;
	p = p->next;
	i++;
    }
    bytes = pages * repeat;
    repeatNotDone = repeat;
    StartTimer();
    while ( repeatNotDone ) {
	repeatNotDone--;
	for ( i = 0; i < pages; i++ ) {
	    dummy += *((volatile u_int32_t*)(addrs[i]));
	}
    }
    StopTimer();
    free( addrs );
    {
	long long Bps;
	Bps = bytes;
	Bps *= 1000000;
	Bps /= TICKS_TO_USEC(BenchTime());
	TE(newResult("readL", repeat, pages, bytes, TICKS_TO_USEC(BenchTime()), Bps ));
    }
    globaldummy += dummy;
    return Success;
}

#if __ppc
#define NOP 0x60000000
#define BLR 0x4e800020
#else
#endif

#if (defined(NOP) && defined(BLR)) && 0
/* Disabled until sync works */
Error testNopSpeed( Page* pIn, int repeat ) {
    int pages = 0;
    unsigned long bytes = 0;
    Address vf;
    void (*pageFunc)(void);
    Page* p;
    int repeatNotDone = repeat;
    p = pIn;
    while ( p ) {
	unsigned long len, i;
	pages++;
	len = p->bytes;
	vf = p->vFirst;
	len = len >> 2;
	for ( i = 0; i < len; i++ ) {
	    *((volatile u_int32_t*)(vf + i)) = NOP;
	}
	*((volatile u_int32_t*)(vf + len - 1)) = BLR;
	// Do something here to sync
	TE(UnmapMemory( __ghs_VirtualMemoryRegionPool, p->vmr ));
	TE(MapMemory( __ghs_VirtualMemoryRegionPool, p->pmr, &(p->vmr), &(p->vFirst), &(p->bytes) ));
	p = p->next;
    }
    StartTimer();
    while ( repeatNotDone ) {
	p = pIn;
	repeatNotDone--;
	while ( p ) {
	    bytes += p->bytes;
	    pageFunc = (void (*)(void))(p->vFirst);
	    pageFunc();
	    p = p->next;
	}
    }
    StopTimer();
    {
	long long Bps;
	Bps = bytes;
	Bps *= 1000000;
	Bps /= TICKS_TO_USEC(BenchTime());
	TE(newResult("nop", repeat, pages, bytes, TICKS_TO_USEC(BenchTime()), Bps ));
    }
    return Success;
}
#else
Error testNopSpeed( Page* pIn, int repeat ) {
    return Success;
}
#endif

#ifndef MAX_PAGES
int MAX_PAGES = 128;
#endif

Error iMain( int argc, char** argv ) {
    Error e;
    Page* root = NULL;
    int pages = 0;
    int repeat;
    while ( ((e = getPage( &root )) == Success) && (pages < MAX_PAGES) ) {
	pages++;
	repeat = 512 / pages;
	if ( repeat == 0 ) repeat = 1;
	TE(testReadSpeed1(root,repeat));
	TE(testReadSpeed2(root,repeat));
	TE(testReadSpeed4(root,repeat));
	TE(testWriteSpeed1(root,repeat));
	TE(testWriteSpeed2(root,repeat));
	TE(testWriteSpeed4(root,repeat));
	TE(testNopSpeed(root,repeat));
	TE(testWriteLatency(root,repeat));
	TE(testReadLatency(root,repeat));
    }
    return e;
}
int main( int argc, char** argv ) {
    Error e;
    StartTimer(); sleep(1); StopTimer();
    my_printf("one second equals 1000000 usec, right ?: %d\n", TICKS_TO_USEC(BenchTime()) );
    my_printf("MAX_PAGES %d\n", MAX_PAGES );
    my_printf("begin memory bandwidth benchmark\n");
    e = iMain( argc, argv );
    printAllResults();
    my_printf("end   memory bandwidth benchmark\n");
    return e != Success;
}
